
/* module.c: Sony PH Key Driver.
 *
 * Copyright 2016 Sony Corporation.
 * Author: Sony Corporation.
 *
 * This package is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/miscdevice.h>
#include <linux/input.h>
#include <asm/uaccess.h>
#include <mach/mt_gpio.h>
#include <mach/eint.h>
#include <mach/icx_pm_helper.h>
#include <mt_auxadc.h>
#include <icx_key.h>
#include "ph_key.h"

#define DRV_NAME					"ph_key"

#define DBGTRACE(fmt, ...)			//
//#define DBGTRACE(fmt, ...)			printk("\e[32m[" DRV_NAME "] "       fmt "\e[0m\n", ## __VA_ARGS__)
#define ERRTRACE(fmt, ...)			printk("\e[31m[" DRV_NAME "] error " fmt "\e[0m\n", ## __VA_ARGS__)


	enum {
		SCAN_INTERVAL_CHAT_GUARD	=	10,
		SCAN_INTERVAL_UNPUSH_DETECT	=	50,
		SCAN_INTERVAL_POLLING		=	100,
		SCAN_WAKEUP_DELAY			=	500,
		ADKEY_CHAT_TRIAL_CNT		=	4,
		GPIOKEY_CHAT_TRIAL_CNT		=	10
	} ;

	/* EINT */
	struct eint {
		u16							flags ;							/* フラグ    */
		u16							num ;							/* EINT 番号 */
	} ;

	/* AD キー */
	struct akey {
		int							state ;							/* 状態        */
		int							state_old[2] ;					/* 以前の状態  */
		struct phkey_akeyinfo		const *p_info ;					/* AD キー情報 */
	} ;

	/* GPIO キー */
	struct gkey {
		u16							eint_num ;						/* EINT 番号  */
		u16							state ;							/* 状態       */
		u16							state_old[2] ;					/* 以前の状態 */
		u16							neg ;							/* 反転値     */
		u16							port ;							/* ポート番号 */
		u16							key_code ;						/* キーコード */
	} ;

	/* プライベートデータ */
	struct priv {
		struct miscdevice			misc ;							/* MISC デバイス           */
		struct input_dev			*p_input ;						/* 入力デバイス            */
		u8							eint_enable ;					/* EINT 割り込み有効フラグ */
		u8							suspend ;						/* サスペンド状態フラグ    */
		u8							key_count ;						/* キーコード数            */
		u32							*p_key_code ;					/* キーコードリスト        */

		struct {
			struct delayed_work		work ;							/* ワーク                  */
			u8						count ;							/* キー数                  */
			u8						eint_count ;					/* EINT 数                 */
			u8						polling ;						/* ポーリングフラグ        */
			struct akey 			*p_key ;						/* AD キー                 */
			struct eint				*p_eint ;						/* EINT                    */
		} akey ;

		struct {
			struct delayed_work		work ;							/* ワーク                  */
			u8						count ;							/* キー数                  */
			u8						eint_count ;					/* EINT 数                 */
			u8						polling ;						/* ポーリングフラグ        */
			struct gkey 			*p_key ;						/* GPIO キー               */
			struct eint				*p_eint ;						/* EINT                    */
		} gkey ;
	} ;


static struct priv					*m_p_priv ;


/* -------------------------------------------------------------------------------- */
/*   EINT 設定                                                                      */
/*                                                                                  */
/*   IN : p_eint = EINT 格納アドレス                                                */
/*        count  = EINT 数                                                          */
/*        num    = EINT 番号                                                        */
/*   OUT: EINT 数                                                                   */
/* -------------------------------------------------------------------------------- */
static inline u8 set_eint(struct eint *p_eint, u8 count, u16 flags, u16 num)
{
	uint	index ;


	for(index=0; index<count; index++) {
		if( p_eint[index].num == num ) {
			return count ;
		}
	}

	p_eint[count].flags = flags ;
	p_eint[count].num   = num ;

	return (count+1) ;
}


/* -------------------------------------------------------------------------------- */
/*   EINT マスク                                                                    */
/*                                                                                  */
/*   IN : enable  = 0:無効, 1:有効                                                  */
/*        polling = ポーリング                                                      */
/*        p_eint  = EINT                                                            */
/*        count   = EINT 数                                                         */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void mask_eint(uint enable, uint polling, struct eint *p_eint, uint count)
{
	struct eint		*p_eint_end ;
	uint			unmask ;


	p_eint_end = &p_eint[count] ;

	while( p_eint != p_eint_end ) {
		if( polling ) {
			unmask = (!enable && (p_eint->flags&PHKEY_FLAG_WAKEUP)) ;
		} else {
			unmask = ( enable || (p_eint->flags&PHKEY_FLAG_WAKEUP)) ;
		}

		if( unmask ) {
			DBGTRACE("unmask = %d", p_eint->num) ;
			mt_eint_unmask( p_eint->num ) ;
		} else {
			DBGTRACE("mask = %d", p_eint->num) ;
			mt_eint_mask( p_eint->num ) ;
		}
		p_eint++ ;
	}
}


/* -------------------------------------------------------------------------------- */
/*   EINT の有効・無効                                                              */
/*                                                                                  */
/*   IN : p_priv = プライベートデータ                                               */
/*        enable = 0:無効, 1:有効                                                   */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void enable_eint(struct priv *p_priv, uint enable)
{
	if( p_priv->eint_enable != enable ) {
		p_priv->eint_enable  = enable ;
		mask_eint(enable, p_priv->akey.polling, &p_priv->akey.p_eint[0], p_priv->akey.eint_count) ;
		mask_eint(enable, p_priv->gkey.polling, &p_priv->gkey.p_eint[0], p_priv->gkey.eint_count) ;
	}
}


/* -------------------------------------------------------------------------------- */
/*   AD キー情報を設定                                                              */
/*                                                                                  */
/*   IN : p_priv     = プライベートデータ                                           */
/*        p_key_info = AD キー情報                                                  */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void set_akey_info(struct priv *p_priv, struct phkey_akeyinfo const *p_key_info)
{
	uint	index, count, eint_count ;


	eint_count = 0 ;

	for(index=0; index<p_priv->akey.count; index++,p_key_info++) {
		p_priv->akey.p_key[index].state  = p_key_info->count ;
		p_priv->akey.p_key[index].p_info = p_key_info ;

		if( !(p_key_info->flags&PHKEY_FLAG_POLLING) ) {
			p_priv->akey.eint_count = set_eint(&p_priv->akey.p_eint[0], p_priv->akey.eint_count, p_key_info->flags, p_key_info->eint_num) ;
			eint_count++ ;
		}

		for(count=0; count<p_key_info->count; count++) {
			p_priv->p_key_code[p_priv->key_count++] = p_key_info->p_data[count].key_code ;
			__set_bit(p_key_info->p_data[count].key_code, p_priv->p_input->keybit) ;
		}
	}

	p_priv->akey.polling = (eint_count != p_priv->akey.count) ;
}


/* -------------------------------------------------------------------------------- */
/*   GPIO キー情報を設定                                                            */
/*                                                                                  */
/*   IN : p_priv     = プライベートデータ                                           */
/*        p_key_info = GPIO キー情報                                                */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void set_gkey_info(struct priv *p_priv, struct phkey_gkeyinfo const *p_key_info)
{
	uint	index, eint_count ;


	eint_count = 0 ;

	for(index=0; index<p_priv->gkey.count; index++,p_key_info++) {
		p_priv->gkey.p_key[index].eint_num = p_key_info->eint_num ;
		p_priv->gkey.p_key[index].neg      = ((p_key_info->flags & PHKEY_FLAG_HIGH) == 0) ;
		p_priv->gkey.p_key[index].port     = p_key_info->port ;
		p_priv->gkey.p_key[index].key_code = p_key_info->key_code ;

		if( !(p_key_info->flags&PHKEY_FLAG_POLLING) ) {
			p_priv->gkey.eint_count = set_eint(&p_priv->gkey.p_eint[0], p_priv->gkey.eint_count, p_key_info->flags, p_key_info->eint_num) ;
			eint_count++ ;
		}

		p_priv->p_key_code[p_priv->key_count++] = p_priv->gkey.p_key[index].key_code ;
		__set_bit(p_priv->gkey.p_key[index].key_code, p_priv->p_input->keybit) ;
	}

	p_priv->gkey.polling = (eint_count != p_priv->gkey.count) ;
}


/* -------------------------------------------------------------------------------- */
/*   AD キーの状態を更新                                                            */
/*                                                                                  */
/*   IN : p_priv = プライベートデータ                                               */
/*   OUT: -1    != 更新数                                                           */
/*        -1    == 異常終了                                                         */
/* -------------------------------------------------------------------------------- */
static int update_akey_state( struct priv *p_priv )
{
	struct akey					*p_akey, *p_akey_end ;
	struct phkey_akeydata		const *p_akey_data ;
	struct phkey_akeyinfo		const *p_akey_info ;
	int							result, index, count, value, data[4] ;


	p_akey     = &p_priv->akey.p_key[0] ;
	p_akey_end = &p_priv->akey.p_key[p_priv->akey.count] ;
	count      = 0 ;

	do {
		p_akey_info          = p_akey->p_info ;
		p_akey->state_old[1] = p_akey->state_old[0] ;
		result               = IMM_GetOneChannelValue(p_akey_info->channel, &data[0], &value) ;
		if( result ) {
			return -1 ;
		}

		p_akey_data = p_akey_info->p_data ;
		for(index=0; index<p_akey_info->count; index++,p_akey_data++) {
			if( (u16 )(value-p_akey_data->min) <= p_akey_data->range ) {
				break ;
			}
		}
		p_akey->state_old[0] = index ;

		count += (p_akey->state_old[0] == p_akey->state_old[1]) ;
	} while( ++p_akey != p_akey_end ) ;

	return count ;
}


/* -------------------------------------------------------------------------------- */
/*   AD キーのイベントを更新                                                        */
/*                                                                                  */
/*   IN : p_priv = プライベートデータ                                               */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void update_akey_event( struct priv *p_priv )
{
	struct akey				*p_akey, *p_akey_end ;
	struct phkey_akeyinfo	const *p_akey_info ;
	int						count ;


	p_akey     = &p_priv->akey.p_key[0] ;
	p_akey_end = &p_akey[p_priv->akey.count] ;
	count      = 0 ;

	do {
		p_akey_info = p_akey->p_info ;

		if( p_akey->state != p_akey->state_old[0] ) {
			if( p_akey->state != p_akey_info->count ) {
				input_event(p_priv->p_input, EV_KEY, p_akey_info->p_data[p_akey->state].key_code, 0) ;
				input_sync( p_priv->p_input ) ;
			}

			p_akey->state = p_akey->state_old[0] ;
			if( p_akey->state != p_akey_info->count ) {
				input_event(p_priv->p_input, EV_KEY, p_akey_info->p_data[p_akey->state].key_code, 1) ;
				input_sync( p_priv->p_input ) ;
			}
		}

		count += (p_akey->state != p_akey_info->count) ;
	} while( ++p_akey != p_akey_end ) ;

	if( count ) {					/* 押下中 */
		schedule_delayed_work(&p_priv->akey.work, msecs_to_jiffies( SCAN_INTERVAL_UNPUSH_DETECT )) ;
	} else
	if( p_priv->akey.polling ) {	/* ポーリング */
		schedule_delayed_work(&p_priv->akey.work, msecs_to_jiffies( SCAN_INTERVAL_POLLING )) ;
	} else {
		DBGTRACE("ad - stop") ;
	}
}


/* -------------------------------------------------------------------------------- */
/*   GPIO キーの状態を更新                                                          */
/*                                                                                  */
/*   IN : p_priv = プライベートデータ                                               */
/*   OUT: 非更新数                                                                  */
/* -------------------------------------------------------------------------------- */
static int update_gkey_state( struct priv *p_priv )
{
	struct gkey		*p_gkey, *p_gkey_end ;
	int				count ;


	p_gkey     = &p_priv->gkey.p_key[0] ;
	p_gkey_end = &p_gkey[p_priv->gkey.count] ;
	count      = 0 ;

	do {
		p_gkey->state_old[1] = p_gkey->state_old[0] ;
		p_gkey->state_old[0] = mt_get_gpio_in( p_gkey->port ) ^ p_gkey->neg ;

		count += (p_gkey->state_old[0] == p_gkey->state_old[1]) ;
	} while( ++p_gkey != p_gkey_end ) ;

	return count ;
}


/* -------------------------------------------------------------------------------- */
/*   GPIO キーのイベントを更新                                                      */
/*                                                                                  */
/*   IN : p_priv = プライベートデータ                                               */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void update_gkey_event( struct priv *p_priv )
{
	struct gkey		*p_gkey, *p_gkey_end ;
	int				count ;


	p_gkey     = &p_priv->gkey.p_key[0] ;
	p_gkey_end = &p_gkey[p_priv->gkey.count] ;
	count      = 0 ;

	do {
		if( p_gkey->state != p_gkey->state_old[0] ) {
			p_gkey->state  = p_gkey->state_old[0] ;
			input_event(p_priv->p_input, EV_KEY, p_gkey->key_code, p_gkey->state) ;
			input_sync( p_priv->p_input ) ;
		}

		count += p_gkey->state ;
	} while( ++p_gkey != p_gkey_end ) ;

	if( count ) {					/* 押下中 */
		schedule_delayed_work(&p_priv->gkey.work, msecs_to_jiffies( SCAN_INTERVAL_UNPUSH_DETECT )) ;
	} else
	if( p_priv->gkey.polling ) {	/* ポーリング */
		schedule_delayed_work(&p_priv->gkey.work, msecs_to_jiffies( SCAN_INTERVAL_POLLING )) ;
	} else {
		DBGTRACE("gpio - stop") ;
	}
}


/* -------------------------------------------------------------------------------- */
/*   AD キー割り込み処理                                                            */
/*                                                                                  */
/*   IN : なし                                                                      */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void akey_isr( void )
{
	struct priv		*p_priv=m_p_priv ;


	DBGTRACE("akey_isr") ;

	if( !p_priv->suspend ) {
		schedule_delayed_work(&p_priv->akey.work, 0) ;
	}
}


/* -------------------------------------------------------------------------------- */
/*   GPIO キー割り込み処理                                                          */
/*                                                                                  */
/*   IN : なし                                                                      */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void gkey_isr( void )
{
	struct priv		*p_priv=m_p_priv ;


	DBGTRACE("gkey_isr") ;

	if( !p_priv->suspend ) {
		schedule_delayed_work(&p_priv->gkey.work, 0) ;
	}
}


/* -------------------------------------------------------------------------------- */
/*   AD キーのワークキュー処理                                                      */
/*                                                                                  */
/*   IN : p_work = ワーク                                                           */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void akey_workqueue_proc( struct work_struct *p_work )
{
	struct priv		*p_priv=container_of(p_work, struct priv, akey.work.work) ;
	int				result, retry ;


	DBGTRACE("ad - start") ;

	result = update_akey_state( p_priv ) ;
	if( result < 0 ) {
		schedule_delayed_work(&p_priv->akey.work, msecs_to_jiffies( SCAN_INTERVAL_POLLING )) ;
		return ;
	}

	for(retry=1; retry<ADKEY_CHAT_TRIAL_CNT; retry++) {
		msleep_interruptible( SCAN_INTERVAL_CHAT_GUARD ) ;

		result = update_akey_state( p_priv ) ;
		if( result < 0 ) {
			break ;
		}

		if( result == p_priv->akey.count ) {
			update_akey_event( p_priv ) ;
			enable_eint(p_priv, 1) ;
			return ;
		}
	}

	schedule_delayed_work(&p_priv->akey.work, msecs_to_jiffies( SCAN_INTERVAL_POLLING )) ;
}


/* -------------------------------------------------------------------------------- */
/*   GPIO キーのワークキュー処理                                                    */
/*                                                                                  */
/*   IN : p_work = ワーク                                                           */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void gkey_workqueue_proc( struct work_struct *p_work )
{
	struct priv		*p_priv=container_of(p_work, struct priv, gkey.work.work) ;
	int				result, retry ;


	DBGTRACE("gpio - start") ;

	update_gkey_state( p_priv ) ;

	for(retry=1; retry<GPIOKEY_CHAT_TRIAL_CNT; retry++) {
		result = update_gkey_state( p_priv ) ;

		if( result == p_priv->gkey.count ) {
			update_gkey_event( p_priv ) ;
			return ;
		}
	}

	schedule_delayed_work(&p_priv->gkey.work, msecs_to_jiffies( SCAN_INTERVAL_POLLING )) ;
}


/* -------------------------------------------------------------------------------- */
/*   開く                                                                           */
/*                                                                                  */
/*   IN : p_inode = i ノード                                                        */
/*        p_file  = ファイル                                                        */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int ph_key_open(struct inode *p_inode, struct file *p_file)
{
	p_file->private_data = container_of(p_file->private_data, struct priv, misc) ;

	return 0 ;
}


/* -------------------------------------------------------------------------------- */
/*  書き込み                                                                        */
/*                                                                                  */
/*  IN : p_file = ファイル                                                          */
/*       p_buff = 書き込みバッファ                                                  */
/*       size   = 書き込みサイズ                                                    */
/*       p_pos  = 位置格納アドレス                                                  */
/*  OUT: 書き込みサイズ                                                             */
/* -------------------------------------------------------------------------------- */
static ssize_t ph_key_write(struct file *p_file, const char __user *p_buff, size_t size, loff_t *p_pos)
{
	struct priv		*p_priv=p_file->private_data ;
	char			buff[32], *p_end ;
	u32				index, value ;


	if( size >= sizeof( buff ) ) {
		return -EINVAL ;
	}

	if( copy_from_user(&buff[0], &p_buff[0], size) ) {
		return -EFAULT ;
	}

	/* インデックス */
	index = simple_strtoul(&buff[0], &p_end, 0) ;
	if( (index>=p_priv->key_count) || (*p_end!=' ') ) {
		return -EINVAL ;
	}

	/* 設定値 */
	value = simple_strtoul(&p_end[1], NULL, 0) ;
	if( value > 1 ) {
		return -EINVAL ;
	}

	input_event(p_priv->p_input, EV_KEY, p_priv->p_key_code[index], value) ;
	input_sync( p_priv->p_input ) ;

	return size ;
}


/* -------------------------------------------------------------------------------- */
/*   ioctl                                                                          */
/*                                                                                  */
/*   IN : p_file = ファイル                                                         */
/*        cmd    = コマンド                                                         */
/*        arg    = 引数                                                             */
/*   OUT: 0                                                                         */
/*        キーコード数                                                              */
/*        -EFAULT                                                                   */
/*        -EINVAL                                                                   */
/* -------------------------------------------------------------------------------- */
static long ph_key_ioctl(struct file *p_file, unsigned int cmd, unsigned long arg)
{
	struct priv		*p_priv=p_file->private_data ;


	switch( cmd ) {
		case ICX_KEY_IOCTL_CMD_GETCOUNT:
		{
			return p_priv->key_count ;
		}

		case ICX_KEY_IOCTL_CMD_GETINFO:
		{
			if( copy_to_user((void __user *)arg, &p_priv->p_key_code[0], sizeof( u32 )*p_priv->key_count) ) {
				return -EFAULT ;
			}
			break ;
		}

		default:
			return -EINVAL ;
	}

	return 0 ;
}


static const struct file_operations		ph_key_fops={
											.owner          = THIS_MODULE,
											.open           = ph_key_open,
											.write          = ph_key_write,
											.unlocked_ioctl = ph_key_ioctl,
											.compat_ioctl   = ph_key_ioctl,
										} ;


/* -------------------------------------------------------------------------------- */
/*   プローブ処理                                                                   */
/*                                                                                  */
/*   IN : p_dev = プラットフォームデバイス                                          */
/*   OUT: 0                                                                         */
/*        -ENOMEM                                                                   */
/* -------------------------------------------------------------------------------- */
static int ph_key_probe( struct platform_device *p_dev )
{
	struct phkey_config		const *p_config=p_dev->dev.platform_data ;
	struct phkey_akeyinfo	const *p_akey_info ;
	struct eint				*p_eint ;
	struct priv				*p_priv ;
	int						result ;
	u16						const *p_key_code ;
	uint					index, count, size ;

	/* キーコード数取得 */
	p_akey_info = &p_config->akey.p_info[0] ;
	count       = p_config->count + p_config->gkey.count ;
	for(index=0; index<p_config->akey.count; index++,p_akey_info++) {
		count += p_akey_info->count ;
	}

	/* プライベートデータ確保 */
	size   =  sizeof( struct priv )                                                   +
	         ((sizeof( struct akey ) + sizeof( struct eint )) * p_config->akey.count) +
	         ((sizeof( struct gkey ) + sizeof( struct eint )) * p_config->gkey.count) +
	         ( sizeof( u32 )                                  * count) ;
	p_priv = kzalloc(size, GFP_KERNEL) ;
	if( p_priv == NULL ) {
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		return -ENOMEM ;
	}
	p_priv->eint_enable = 1 ;
	p_priv->akey.count  = p_config->akey.count ;
	p_priv->gkey.count  = p_config->gkey.count ;
	p_priv->akey.p_key  = (struct akey *)&p_priv[1] ;
	p_priv->gkey.p_key  = (struct gkey *)&p_priv->akey.p_key[p_priv->akey.count] ;
	p_priv->akey.p_eint = (struct eint *)&p_priv->gkey.p_key[p_priv->gkey.count] ;
	p_priv->gkey.p_eint = (struct eint *)&p_priv->akey.p_eint[p_priv->akey.count] ;
	p_priv->p_key_code  = (u32         *)&p_priv->gkey.p_eint[p_priv->gkey.count] ;

	INIT_DELAYED_WORK(&p_priv->akey.work, akey_workqueue_proc) ;
	INIT_DELAYED_WORK(&p_priv->gkey.work, gkey_workqueue_proc) ;

	/* MISC デバイス作成 */
	p_priv->misc.minor = MISC_DYNAMIC_MINOR ;
	p_priv->misc.name  = DRV_NAME ;
	p_priv->misc.fops  = &ph_key_fops ;
	result             = misc_register( &p_priv->misc ) ;
	if( result ) {
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		goto ph_key_probe_f00 ;
	}

	/* 入力デバイス確保 */
	p_priv->p_input = input_allocate_device() ;
	if( p_priv->p_input == NULL ) {
		result = -ENOMEM ;
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		goto ph_key_probe_e00 ;
	}

	/* 仮想キーコード設定 */
	p_key_code = &p_config->p_key_code[0] ;
	for(index=0; index<p_config->count; index++,p_key_code++) {
		p_priv->p_key_code[p_priv->key_count++] = *p_key_code ;
		__set_bit(*p_key_code, p_priv->p_input->keybit) ;
	}

	set_akey_info(p_priv, p_config->akey.p_info) ;
	set_gkey_info(p_priv, p_config->gkey.p_info) ;

	DBGTRACE("alloc size = %d - %d", size, (int )((u8 *)&p_priv->p_key_code[p_priv->key_count]-(u8 *)p_priv)) ;

	/* 入力デバイス登録 */
	__set_bit(EV_KEY, p_priv->p_input->evbit) ;
	p_priv->p_input->name       = DRV_NAME ;
	p_priv->p_input->phys       = DRV_NAME ;
	p_priv->p_input->id.bustype = BUS_HOST ;
	p_priv->p_input->id.vendor  = 0x0001 ;
	p_priv->p_input->id.product = 0x0001 ;
	p_priv->p_input->id.version = 0x0100 ;
	p_priv->p_input->dev.parent = &p_dev->dev ;
	result                      = input_register_device( p_priv->p_input ) ;
	if( result ) {
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		goto ph_key_probe_d00 ;
	}

	platform_set_drvdata(p_dev, p_priv) ;

	m_p_priv = p_priv ;

	/* AD キー割り込み */
	p_eint = &p_priv->akey.p_eint[0] ;
	for(index=0; index<p_priv->akey.eint_count; index++,p_eint++) {
		if( p_eint->flags & PHKEY_FLAG_HIGH ) {
			mt_eint_registration(p_eint->num, EINTF_TRIGGER_RISING,  akey_isr, 1) ;
		} else {
			mt_eint_registration(p_eint->num, EINTF_TRIGGER_FALLING, akey_isr, 1) ;
		}
		if( p_priv->akey.polling ) {
			mt_eint_mask( p_eint->num ) ;
		}
	}

	/* GPIO キー割り込み */
	p_eint = &p_priv->gkey.p_eint[0] ;
	for(index=0; index<p_priv->gkey.eint_count; index++,p_eint++) {
		if( p_eint->flags & PHKEY_FLAG_HIGH ) {
			mt_eint_registration(p_eint->num, EINTF_TRIGGER_RISING,  gkey_isr, 1) ;
		} else {
			mt_eint_registration(p_eint->num, EINTF_TRIGGER_FALLING, gkey_isr, 1) ;
		}
		if( p_priv->gkey.polling ) {
			mt_eint_mask( p_eint->num ) ;
		}
	}

	if( p_priv->akey.polling ) {
		schedule_delayed_work(&p_priv->akey.work, msecs_to_jiffies( SCAN_INTERVAL_POLLING )) ;
	}

	if( p_priv->gkey.polling ) {
		schedule_delayed_work(&p_priv->gkey.work, msecs_to_jiffies( SCAN_INTERVAL_POLLING )) ;
	}

	return 0 ;


ph_key_probe_d00:
	input_free_device( p_priv->p_input ) ;
ph_key_probe_e00:
	misc_deregister( &p_priv->misc ) ;
ph_key_probe_f00:
	kfree( p_priv ) ;

	return result ;
}


/* -------------------------------------------------------------------------------- */
/*   削除処理                                                                       */
/*                                                                                  */
/*   IN : p_dev = プラットフォームデバイス                                          */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int ph_key_remove( struct platform_device *p_dev )
{
	struct priv		*p_priv=platform_get_drvdata( p_dev ) ;
	struct eint		*p_eint ;
	uint			index ;

	/* 割り込み解除 */
	p_eint = &p_priv->akey.p_eint[0] ;
	for(index=0; index<p_priv->akey.eint_count; index++,p_eint++) {
		mt_eint_registration(p_eint->num, EINTF_TRIGGER_RISING, NULL, 1) ;
	}

	p_eint = &p_priv->gkey.p_eint[0] ;
	for(index=0; index<p_priv->gkey.eint_count; index++,p_eint++) {
		mt_eint_registration(p_eint->num, EINTF_TRIGGER_RISING, NULL, 1) ;
	}

	cancel_delayed_work_sync( &p_priv->akey.work ) ;
	cancel_delayed_work_sync( &p_priv->gkey.work ) ;

	input_unregister_device( p_priv->p_input ) ;

	misc_deregister( &p_priv->misc ) ;

	kfree( p_priv ) ;

	return 0 ;
}


/* -------------------------------------------------------------------------------- */
/*   サスペンド処理                                                                 */
/*                                                                                  */
/*   IN : p_dev = プラットフォームデバイス                                          */
/*        state = 状態                                                              */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int ph_key_suspend(struct platform_device *p_dev, pm_message_t state)
{
	struct priv		*p_priv=platform_get_drvdata( p_dev ) ;
	struct gkey		*p_gkey ;
	uint			index ;


	DBGTRACE("suspend") ;

	p_priv->suspend = 1 ;

	enable_eint(p_priv, 0) ;

	if( p_priv->akey.count ) {
		cancel_delayed_work_sync( &p_priv->akey.work ) ;
	}

	if( p_priv->gkey.count ) {
		cancel_delayed_work_sync( &p_priv->gkey.work ) ;
	}

	/* 押下キー解除 */
	p_gkey = &p_priv->gkey.p_key[0] ;
	for(index=0; index<p_priv->gkey.count; index++,p_gkey++) {
		if( p_gkey->state ) {
			input_event(p_priv->p_input, EV_KEY, p_gkey->key_code, 0) ;
			input_sync( p_priv->p_input ) ;
			p_gkey->state = 0 ;
		}
		p_gkey->state_old[0] = 0 ;
		p_gkey->state_old[1] = 0 ;
	}

	return 0 ;
}


/* -------------------------------------------------------------------------------- */
/*   レジューム処理                                                                 */
/*                                                                                  */
/*   IN : p_dev = プラットフォームデバイス                                          */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int ph_key_resume( struct platform_device *p_dev )
{
	struct priv		*p_priv=platform_get_drvdata( p_dev ) ;
	struct gkey		*p_gkey ;
	uint			index ;


	DBGTRACE("resume") ;

	/* 割り込みに使用しているキー押下 */
	if( !p_priv->akey.polling ) {
		p_gkey = &p_priv->gkey.p_key[0] ;
		for(index=0; index<p_priv->gkey.count; index++,p_gkey++) {
			if( icx_pm_helper_spm_stat.eint_sta[p_gkey->eint_num/32] & (1<<(p_gkey->eint_num&31)) ) {
				p_gkey->state        = 1 ;
				p_gkey->state_old[0] = 1 ;
				p_gkey->state_old[1] = 1 ;
				input_event(p_priv->p_input, EV_KEY, p_gkey->key_code, 1) ;
				input_sync( p_priv->p_input ) ;
			}
		}
	}

	enable_eint(p_priv, 1) ;

	p_priv->suspend = 0 ;

	if( p_priv->akey.count ) {
		schedule_delayed_work(&p_priv->akey.work, msecs_to_jiffies( SCAN_WAKEUP_DELAY )) ;
	}

	if( p_priv->gkey.count ) {
		schedule_delayed_work(&p_priv->gkey.work, msecs_to_jiffies( SCAN_WAKEUP_DELAY )) ;
	}

	return 0 ;
}


static struct platform_driver		ph_key_platform_driver={
										.probe     = ph_key_probe,
										.remove    = ph_key_remove,
										.suspend   = ph_key_suspend,
										.resume    = ph_key_resume,
										.driver    = {
											.name  = DRV_NAME,
											.owner = THIS_MODULE,
										},
									} ;


module_platform_driver( ph_key_platform_driver ) ;


MODULE_DESCRIPTION("Sony PH Key Driver");
MODULE_AUTHOR("Sony Corporation");
MODULE_LICENSE("GPL") ;
MODULE_ALIAS("Sony PH Key Driver") ;
MODULE_VERSION( "v0.01" ) ;
